/*
 * FPassword.java
 *
 * Copyright  2004 Wayne Grant
 * waynedgrant@hotmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * (This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package info.waynegrant.ant;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/**
 * Dialog used by the password Ant task to collect a password and optionally
 * a username.
 *
 * @author Wayne Grant
 */
class FPassword extends JFrame // Deliberately package private
{
    /** Panel to hold password details controls */
    private JPanel m_jpDetails;

    /** Label for description */
    private JLabel m_jlDescription;

    /** Label for username */
    private JLabel m_jlUsername;

    /** Label for password */
    private JLabel m_jlPassword;

    /** Username entry text field */
    private JTextField m_jtfUsername;

    /** Password entry password field */
    private JPasswordField m_jpfPassword;

    /** Panel to hold OK and cancel buttons */
    private JPanel m_jpButtons;

    /** OK button to confirm password entry */
    private JButton m_jbOK;

    /** Cancel to button to cancel password entry */
    private JButton m_jbCancel;

    /** Dialog description */
    private String m_sDescription;

    /** Username prompt */
    private String m_sUsernamePrompt;

    /** Password prompt */
    private String m_sPasswordPrompt;

    /** Stores username */
    private String m_sUsername;

    /** Stores password */
    private char[] m_cPassword;

    /** Key from input map to action map for the cancel button */
    private static final String CANCEL_KEY = "CANCEL_KEY";

    /**
     * Creates a new FPassword dialog to retrieve a password.
     *
     * @param sDescription Dialog description
     * @param sPasswordPrompt Password prompt
     */
    public FPassword(String sDescription, String sPasswordPrompt)
    {
        super(sDescription);
        m_sDescription = sDescription;
        m_sPasswordPrompt = sPasswordPrompt;
        initComponents();
    }

    /**
     * Creates a new FPassword dialog to retrieve a username and password.
     *
     * @param sDescription Dialog description
     * @param sUsernamePrompt Username prompt
     * @param sPasswordPrompt Password prompt
     */
    public FPassword(String sDescription, String sUsernamePrompt, String sPasswordPrompt)
    {
        super(sDescription);
        m_sDescription = sDescription;
        m_sUsernamePrompt = sUsernamePrompt;
        m_sPasswordPrompt = sPasswordPrompt;
        initComponents();
    }

    /**
     * Initialise the dialog's GUI components.
     */
    private void initComponents()
    {
        getContentPane().setLayout(new BorderLayout());

        m_jlDescription = new JLabel(m_sDescription);
        m_jlDescription.setForeground(Color.red);
        GridBagConstraints gbc_jlDescription = new GridBagConstraints();
        gbc_jlDescription.gridx = 0;
        gbc_jlDescription.gridy = 0;
        gbc_jlDescription.gridwidth = 6;
        gbc_jlDescription.gridheight = 1;
        gbc_jlDescription.weightx = 1;
        gbc_jlDescription.insets = new Insets(5, 5, 5, 5);

        // Are username controls required?
        if (m_sUsernamePrompt != null)
        {
            m_jlUsername = new JLabel(m_sUsernamePrompt);
            m_jtfUsername = new JTextField(10);
        }

        m_jlPassword = new JLabel(m_sPasswordPrompt);
        m_jpfPassword = new JPasswordField(10);

        m_jpDetails = new JPanel(new GridBagLayout());
        m_jpDetails.add(m_jlDescription, gbc_jlDescription);

        // Are username controls required
        if (m_sUsernamePrompt != null)
        {
            m_jpDetails.add(m_jlUsername, new GridBagConstraints(0, 1, 3, 1, 1, 0, GridBagConstraints.EAST, 0, new Insets(5, 5, 5, 5), 0, 0));
            m_jpDetails.add(m_jtfUsername, new GridBagConstraints(3, 1, 3, 1, 1, 0, GridBagConstraints.WEST, 0, new Insets(5, 5, 5, 5), 0, 0));
            m_jpDetails.add(m_jlPassword, new GridBagConstraints(0, 2, 3, 1, 1, 0, GridBagConstraints.EAST, 0, new Insets(5, 5, 5, 5), 0, 0));
            m_jpDetails.add(m_jpfPassword, new GridBagConstraints(3, 2, 3, 1, 1, 0, GridBagConstraints.WEST, 0, new Insets(5, 5, 5, 5), 0, 0));
        }
        else
        {
            m_jpDetails.add(m_jlPassword, new GridBagConstraints(0, 1, 3, 1, 1, 0, GridBagConstraints.EAST, 0, new Insets(5, 5, 5, 5), 0, 0));
            m_jpDetails.add(m_jpfPassword, new GridBagConstraints(3, 1, 3, 1, 1, 0, GridBagConstraints.WEST, 0, new Insets(5, 5, 5, 5), 0, 0));
        }

        m_jpDetails.setBorder(new CompoundBorder(new EmptyBorder(5, 5, 5, 5),
                                                 new EtchedBorder()));

        m_jbOK = new JButton("OK");
        m_jbOK.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                okPressed();
            }
        });

        m_jbCancel = new JButton("Cancel");
        m_jbCancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                cancelPressed();
            }
        });
        m_jbCancel.getInputMap(m_jbCancel.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), CANCEL_KEY);
        m_jbCancel.getActionMap().put(CANCEL_KEY, new AbstractAction () {
                                          public void actionPerformed(ActionEvent evt) {
                                              cancelPressed();
                                      }});

        m_jpButtons = new JPanel(new FlowLayout(FlowLayout.CENTER));
        m_jpButtons.add(m_jbOK);
        m_jpButtons.add(m_jbCancel);

        getContentPane().add(m_jpDetails, BorderLayout.CENTER);
        getContentPane().add(m_jpButtons, BorderLayout.SOUTH);

        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent evt) {
                closeDialog();
            }
        });

        setResizable(false);

        getRootPane().setDefaultButton(m_jbOK);

        setIconImage(Toolkit.getDefaultToolkit().createImage(getClass().getResource("images/password.gif")));

        pack();

        setLocationCenter();
    }

    /**
     * Get the username entered into the dialog.
     *
     * @return The username or null if cancel was pressed
     */
    public String getUsername()
    {
        return m_sUsername;
    }

    /**
     * Get the password entered into the dialog.
     *
     * @return The password or null if cancel was pressed
     */
    public char[] getPassword()
    {
        return m_cPassword;
    }

    /**
     * OK button pressed or otherwise activated.
     */
    private void okPressed()
    {
        if (m_jtfUsername != null)
        {
            m_sUsername = m_jtfUsername.getText();
        }
        m_cPassword = m_jpfPassword.getPassword();
        closeDialog();
    }

    /**
     * Cancel button pressed or otherwise activated.
     */
    private void cancelPressed()
    {
        closeDialog();
    }

    /**
     * Close the dialog.
     */
    private void closeDialog()
    {
        setVisible(false);
        dispose();
    }

    /**
     * Set the dialog's position to the center of the desktop.
     */
    public void setLocationCenter()
    {
        Toolkit t = Toolkit.getDefaultToolkit();
        Dimension screenSize = t.getScreenSize();
        int iX = (screenSize.width / 2) - (this.getWidth() / 2);
        int iY = (screenSize.height / 2) - (this.getHeight() / 2);
        setLocation(new Point(iX, iY));
    }
}
