/*
 * PasswordTask.java
 *
 * Copyright  2004 Wayne Grant
 * waynedgrant@hotmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * (This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package info.waynegrant.ant;

import org.apache.tools.ant.*;

/**
 * Ant task to retrieve a password from a user.  Optionally a username
 * may also be retrieved.
 *
 * @author Wayne Grant
 */
public class PasswordTask extends Task
{
    /** Default value for the description attribute */
    private static final String DEFAULT_DESCRIPTION = "Logon";

    /** Default value for the usernameprompt description attribute */
    private static final String DEFAULT_USERNAME_PROMPT = "Username:";

    /** Default value for the passwordprompt attribute */
    private static final String DEFAULT_PASSWORD_PROMPT = "Password:";

    /** Default value for the passwordproperty attribute */
    private static final String DEFAULT_PASSWORD_PROPERTY = "password";

    /** Stores the value of the description attribute */
    private String m_sDescription;

    /** Stores the value of the usernameprompt attribute */
    private String m_sUsernamePrompt;

    /** Stores the value of the passwordprompt attribute */
    private String m_sPasswordPrompt;

    /** Stores the value of the usernameproperty attribute */
    private String m_sUsernameProperty;

    /** Stores the value of the passwordproperty attribute */
    private String m_sPasswordProperty;

    /**
     * Execute the task.
     *
     * @throws BuildException If a problem occurs while executing the task
     */
    public void execute() throws BuildException
    {
        // Setup default values for attributes
        String sDescription = DEFAULT_DESCRIPTION;
        String sUsernamePrompt = DEFAULT_USERNAME_PROMPT;
        String sPasswordPrompt = DEFAULT_PASSWORD_PROMPT;
        String sUsernameProperty = null;
        String sPasswordProperty = DEFAULT_PASSWORD_PROPERTY;

        // Read the attribute - default is used where attribute is not set
        if (m_sDescription != null)
        {
            sDescription = m_sDescription;
        }

        if (m_sUsernamePrompt != null)
        {
            sUsernamePrompt = m_sUsernamePrompt;
        }

        if (m_sPasswordPrompt != null)
        {
            sPasswordPrompt = m_sPasswordPrompt;
        }

        if (m_sUsernameProperty != null)
        {
            sUsernameProperty = m_sUsernameProperty;
        }

        if (m_sPasswordProperty != null)
        {
            sPasswordProperty = m_sPasswordProperty;
        }

        // Spawn password dialog
        FPassword fPassword = null;

        if (sUsernameProperty != null)
        {
            // Username and password to be retrieved
            fPassword = new FPassword(sDescription, sUsernamePrompt, sPasswordPrompt);
        }
        else
        {
            // Only password to be retrieved
            fPassword = new FPassword(sDescription, sPasswordPrompt);
        }
        fPassword.setVisible(true);

        // Wait for user to close dialog
        while (true)
        {
            if (!fPassword.isVisible())
            {
                break;
            }

            try { Thread.sleep(250); } catch (InterruptedException ex) { /* Ignore */ }
        }

        // Retrieve values entered by user
        String sUsername = fPassword.getUsername();
        char[] cPassword = fPassword.getPassword();

        // If the user entered values then set the appropriate username/password properties
        Project project = getProject();

        if (cPassword != null)
        {
            // Password required and entered
            project.setUserProperty(sPasswordProperty, new String(cPassword));
        }

        if ((sUsernameProperty != null) && (sUsername != null))
        {
            // username required and entered
            project.setUserProperty(sUsernameProperty, sUsername);
        }
    }

    /**
     * Set the description attribute.
     *
     * @param sDescription Attribute value
     */
    public void setDescription(String sDescription)
    {
        m_sDescription = sDescription;
    }

    /**
     * Set the usernameprompt attribute.
     *
     * @param sUsernamePrompt Attribute value
     */
    public void setUsernameprompt(String sUsernamePrompt)
    {
        m_sUsernamePrompt = sUsernamePrompt;
    }

    /**
     * Set the passwordprompt attribute.
     *
     * @param sPasswordPrompt Attribute value
     */
    public void setPasswordprompt(String sPasswordPrompt)
    {
        m_sPasswordPrompt = sPasswordPrompt;
    }

    /**
     * Set the usernameproperty attribute.
     *
     * @param sUsernameProperty Attribute value
     */
    public void setUsernameproperty(String sUsernameProperty)
    {
        m_sUsernameProperty = sUsernameProperty;
    }

    /**
     * Set the passwordproperty property.
     *
     * @param sPasswordProperty Attribute value
     */
    public void setPasswordproperty(String sPasswordProperty)
    {
        m_sPasswordProperty = sPasswordProperty;
    }
}
